from datetime import datetime
from zoneinfo import ZoneInfo

import pandas as pd
import pytest

from cosimtlk.simulation.utils import ensure_tz


def test_ensure_tz_timestamp():
    dt = pd.Timestamp("2021-01-01 00:00:00")
    dt = ensure_tz(dt)
    assert dt.tzinfo == ZoneInfo("UTC")


def test_ensure_tz_timestamp_with_tz():
    dt = pd.Timestamp("2021-01-01 00:00:00", tz=ZoneInfo("Europe/Brussels"))
    dt = ensure_tz(dt)
    assert dt.tzinfo == ZoneInfo("Europe/Brussels")


def test_ensure_tz_datetime():
    dt = datetime(2021, 1, 1, 0, 0, 0)  # noqa
    dt = ensure_tz(dt)
    assert dt.tzinfo == ZoneInfo("UTC")


def test_ensure_tz_datetime_with_tz():
    dt = datetime(2021, 1, 1, 0, 0, 0, tzinfo=ZoneInfo("Europe/Brussels"))
    dt = ensure_tz(dt)
    assert dt.tzinfo == ZoneInfo("Europe/Brussels")


if __name__ == "__main__":
    pytest.main()



from __future__ import annotations

import functools
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Generator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cosimtlk.simulation import Simulator


class EntityLogger(logging.LoggerAdapter):
    def __init__(self, logger, entity):
        self.entity = entity
        self.entity_name = entity.name
        super().__init__(logger, extra={"entity": self.entity_name})

    def log(self, level, msg, *args, **kwargs):
        msg = f"{self.entity.ctx.current_datetime}:{self.entity_name}:{msg}"
        self.logger.log(level, msg, *args, **kwargs)


class Entity(metaclass=ABCMeta):
    def __init__(self, name: str, attributes: dict | None = None) -> None:
        """Entity base class that defines the interface for all entities.

        Abstract base class for all entities in the simulation. An entity is
        a component that can be scheduled in the simulation environment.
        It must implement the `processes` method that returns a list of
        processes that should be scheduled.
        The order of the processes in the list is the order in which they
        will be scheduled in case they should fire exactly the same time.

        Args:
            name: Name of the entity for identification purposes.
            attributes: Dictionary of attributes that can be used to store.
        """
        self._name = name
        self._attributes = attributes or {}
        self._context: Simulator | None = None
        self._logger: EntityLogger | None = None

    def __repr__(self):
        """Representation of the entity."""
        return f"<{self.__class__.__name__}: {self.name}>"

    @property
    def name(self) -> str:
        """Name of the entity."""
        return self._name

    @property
    def attributes(self) -> dict:
        """Attributes of the entity."""
        return self._attributes

    @property
    def ctx(self) -> Simulator:
        """Simulation environment."""
        if self._context is None:
            msg = "Entity has not been initialized yet."
            raise RuntimeError(msg)
        return self._context

    def initialize(self, context: Simulator) -> Entity:
        """Initialize the entity in the simulation.

        Returns:
            List of processes that should be scheduled.
        """
        self._context = context
        self._logger = EntityLogger(context.logger, self)
        return self

    @property
    @abstractmethod
    def processes(self) -> list[Callable[[], Generator]]:
        """List of processes that should be scheduled.

        Returns:
            List of processes.
        """
        raise NotImplementedError

    @functools.cached_property
    def log(self) -> EntityLogger:
        return self._logger

    def wait_for(self, duration: int | float = 0):
        """Wait for a given duration.

        Args:
            duration: Duration to wait for.

        Returns:
            Timeout event.
        """
        return self.ctx.env.timeout(duration)

    def wait_until(self, ts: int | float):
        """Wait until a given timestamp.

        Args:
            ts: Timestamp to wait until.

        Returns:
            Timeout event.
        """
        duration = ts - self.ctx.current_timestamp
        return self.wait_for(duration)


class Source(Entity, metaclass=ABCMeta):
    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.generate]

    @abstractmethod
    def interarrival_time(self) -> int | float:
        raise NotImplementedError

    @abstractmethod
    def build_entity(self):
        raise NotImplementedError

    def _generate_interarrival_time(self):
        # if first_creation exists, emit it as the first time, else just use the interarrival_time
        while True:
            yield self.interarrival_time()

    def generate(self):
        for arrival_time in self._generate_interarrival_time():
            timeout = self.wait_for(arrival_time)
            yield timeout
            entity = self.build_entity()
            self.ctx.add_entity(entity)
            self.log.debug(f"created entity={entity} with attributes={entity.attributes}")



import logging
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import numpy as np
from simpy.util import start_delayed

from cosimtlk.models import DateTimeLike
from cosimtlk.simulation.entities import Entity
from cosimtlk.simulation.environment import Environment
from cosimtlk.simulation.utils import ensure_tz


class Simulator:
    def __init__(
        self,
        *,
        initial_time: DateTimeLike,
        entities: list[Entity] | None = None,
        logger: logging.Logger | None = None,
        **kwargs,
    ) -> None:
        """Simulation runner.

        Organizes the entities processes into a discrete event simulation. If multiple entities would run at the same
        time, the order is determined by the order of the entities in the entities list.

        Args:
            initial_time: The initial time of the simulation. Can be either a datetime or a Timestamp.
            entities: A list of entities inside the simulation.
        """
        initial_time = self._parse_datetime(initial_time)
        initial_timestamp = self._dt_to_timestamp(initial_time)
        self.tzinfo: ZoneInfo = initial_time.tzinfo

        # Create simulation environment
        self._environment = Environment(initial_time=initial_timestamp)

        # Add entities to the simulation
        self._current_process = 0
        self._process_delays = np.linspace(0.005, 0.995, 10000)
        self._entities: dict[str, Entity] = {}
        for entity in entities or []:
            self.add_entity(entity)

        # Set up logger
        self._logger = logger or logging.getLogger(__name__)

        # Set additional attributes
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self) -> str:
        return f"<Simulator t={self.current_datetime} entities=[{self._entities.keys()}]>"

    def add_entity(self, entity: Entity) -> "Simulator":
        """Add an entity to the simulation.

        Args:
            entity: The entity to add to the simulation.

        Returns:
            The simulator object.
        """
        if entity.name in self._entities:
            msg = f"Entity with name {entity.name} already exists."
            raise ValueError(msg)
        self._entities[entity.name] = entity

        entity.initialize(self)
        for process in entity.processes:
            start_delayed(self._environment, process(), float(self._process_delays[self._current_process]))
            self._current_process += 1
        return self

    @property
    def entities(self) -> list[Entity]:
        """The entities inside the simulation."""
        return list(self._entities.values())

    def get_entity(self, name: str) -> Entity:
        """Get an entity by name.

        Args:
            name: The name of the entity to get.

        Returns:
            The entity with the given name.
        """
        return self._entities[name]

    def remove_entity(self, name: str) -> None:
        """Remove an entity from the simulation.

        Args:
            name: The name of the entity to remove.

        Raises:
            KeyError: If the entity does not exist.
        """
        del self._entities[name]

    @property
    def logger(self) -> logging.Logger:
        """The logger for the simulation."""
        return self._logger

    @property
    def env(self) -> Environment:
        """The simulation environment."""
        return self._environment

    @property
    def current_timestamp(self) -> int:
        """The current simulation time as a unix timestamp."""
        return int(self._environment.now)

    @property
    def current_datetime(self) -> datetime:
        """The current simulation time as a timezone aware datetime object."""
        return datetime.fromtimestamp(self.current_timestamp, tz=self.tzinfo)

    @staticmethod
    def _parse_datetime(dt: DateTimeLike) -> DateTimeLike:
        return ensure_tz(dt)

    @staticmethod
    def _dt_to_timestamp(dt: DateTimeLike) -> int:
        """Converts a datetime to a unix timestamp."""
        return int(dt.timestamp())

    @staticmethod
    def _td_to_duration(td: timedelta) -> int:
        """Converts a timedelta to a duration in seconds."""
        return int(td.total_seconds())

    def run(self, until: int | datetime, show_progress_bar: bool = True):  # noqa
        """Runs the simulation until the given timestamp.

        Args:
            until: The timestamp until the simulation should run.
            show_progress_bar: Whether to show a progress bar.
        """
        if isinstance(until, datetime):
            if until.tzinfo is not None and until.tzinfo != self.tzinfo:
                msg = f"Until must be in the same timezone as the initial time. {self.tzinfo} != {until.tzinfo}"
                raise ValueError(msg)
            until = ensure_tz(until, default_tz=self.tzinfo)
            until = self._dt_to_timestamp(until)
        else:
            until = int(until)
        self._environment.run(until=until, show_progress_bar=show_progress_bar)

    def run_for(self, duration: int | timedelta, show_progress_bar: bool = True):  # noqa
        """Runs the simulation for the given duration.

        Args:
            duration: The duration for which the simulation should run.
            show_progress_bar: Whether to show a progress bar.
        """
        if isinstance(duration, timedelta):
            duration = self._td_to_duration(duration)
        else:
            duration = int(duration)
        self.run(until=self.current_timestamp + duration, show_progress_bar=show_progress_bar)



from zoneinfo import ZoneInfo

import pandas as pd
from pandas import DataFrame, Series

from cosimtlk.models import DateTimeLike, FMUInputType
from cosimtlk.simulation.utils import ensure_tz


class ObservationStore:
    def __init__(self) -> None:
        self._db: dict[str, Series] = {}

    @classmethod
    def with_history(cls, **history: Series) -> "ObservationStore":
        store = cls()
        store.store_history(**history)
        return store

    def store_history(self, **history: Series):
        for key, history_ in history.items():
            if not isinstance(history_.index, pd.DatetimeIndex):
                msg = f"The index of the history of '{key}' is not a DatetimeIndex."
                raise ValueError(msg)

            history_.rename_axis("timestamp", inplace=True)
            if history_.index.tz is None:
                history_.index = history_.index.tz_localize(ZoneInfo("UTC"))

            self._db[key] = history_

    def store_observation(self, key: str, value: FMUInputType, ts: DateTimeLike) -> None:
        ts = ensure_tz(ts)
        if key not in self._db:
            self._db[key] = pd.Series(name=key, index=[ts], data=[value]).rename_axis("timestamp")
        self._db[key][ts] = value

    def store_observations(self, ts: DateTimeLike, **values: FMUInputType) -> None:
        for key, value in values.items():
            self.store_observation(key, value, ts)

    def get_observations(
        self,
        key: str,
        start: DateTimeLike | None = None,
        end: DateTimeLike | None = None,
        limit: int | None = None,
    ) -> Series:
        start = ensure_tz(start) if start is not None else None
        end = ensure_tz(end) if end is not None else None

        obs = self._db[key].loc[start:end]
        if limit is not None:
            obs = obs.tail(limit)
        return obs

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self._db)


class ScheduleStore:
    def __init__(self) -> None:
        self._db: dict[str, DataFrame] = {}

    def store_history(self, **history: DataFrame) -> None:
        for key, schedule in history.items():
            if schedule.timestamp.dt.tz is None:
                schedule.timestamp = schedule.timestamp.dt.tz_localize(ZoneInfo("UTC"))

            if schedule.cutoff_date.dt.tz is None:
                schedule.cutoff_date = schedule.cutoff_date.dt.tz_localize(ZoneInfo("UTC"))

            self._db[key] = schedule

    @classmethod
    def with_history(cls, **history: DataFrame) -> "ScheduleStore":
        store = cls()
        store.store_history(**history)
        return store

    @staticmethod
    def _create_schedule_df(schedule: DataFrame, cutoff_date: DateTimeLike) -> DataFrame:
        cutoff_date = ensure_tz(cutoff_date)

        schedule = schedule.copy().rename_axis("timestamp").reset_index().assign(cutoff_date=cutoff_date)
        if schedule.timestamp.dt.tz is None:
            schedule.timestamp = schedule.timestamp.dt.tz_localize(ZoneInfo("UTC"))
        return schedule

    def store_schedule(self, key: str, schedule: DataFrame, cutoff_date: DateTimeLike) -> None:
        schedule = self._create_schedule_df(schedule, cutoff_date)
        if key not in self._db:
            self._db[key] = schedule

        self._db[key] = pd.concat([self._db[key], schedule])

    def store_schedules(self, cutoff_date: DateTimeLike, **schedules: DataFrame) -> None:
        for key, schedule in schedules.items():
            self.store_schedule(key, schedule, cutoff_date)

    def get_schedule(
        self,
        key: str,
        made_after: DateTimeLike | None = None,
        made_before: DateTimeLike | None = None,
    ) -> DataFrame:
        made_after = ensure_tz(made_after) if made_after is not None else None
        made_before = ensure_tz(made_before) if made_before is not None else None

        if made_after is None and made_before is None:
            return self._db[key]

        if made_after is None and made_before is not None:
            mask = self._db[key].index <= made_before
        elif made_after is not None and made_before is None:
            mask = self._db[key].index >= made_after
        else:
            mask = (self._db[key].index >= made_after) & (self._db[key].index <= made_before)
        return self._db[key].loc[mask]

    def get_last_schedule(self, key: str, cutoff_date: DateTimeLike) -> DataFrame:
        cutoff_date = ensure_tz(cutoff_date)
        cutoff_dates = self._db[key]["cutoff_date"].unique()
        last_cutoff_time = cutoff_dates[cutoff_dates <= cutoff_date].max()
        schedule = self.get_schedule(key)
        return schedule.loc[schedule.cutoff_date == last_cutoff_time].set_index("timestamp")



from typing import Any

import simpy as sp
from simpy import Event
from simpy.core import EmptySchedule, SimTime, StopSimulation
from simpy.events import URGENT
from tqdm import tqdm


class Environment(sp.Environment):
    def run(
        self,
        until: SimTime | Event | None = None,
        show_progress_bar: bool = True,  # noqa
    ) -> Any | None:
        """Run the environment until the given event or time.

        Args:
            until: The event or time until which the environment should be run.
            show_progress_bar: Whether to show a progress bar.

        Returns:
            The value of the event if it was triggered, otherwise None.
        """
        if until is not None:
            if not isinstance(until, Event):
                # Assume that *until* is a number if it is not None and
                # not an event.  Create a Timeout(until) in this case.
                at: SimTime
                if isinstance(until, int):
                    at = until
                else:
                    at = float(until)

                if at <= self._now:
                    msg = f"until(={at}) must be > the current simulation time."
                    raise ValueError(msg)

                # Schedule the event before all regular timeouts.
                until = Event(self)
                until._ok = True
                until._value = None
                self.schedule(until, URGENT, at - self._now)

            elif until.callbacks is None:
                # Until event has already been processed.
                return until.value

            until.callbacks.append(StopSimulation.callback)

        try:
            if show_progress_bar:
                total = at - self._now
                pbar = tqdm(total=total, desc="Simulation progress", unit="s")
                while True:
                    now = int(self._now)
                    self.step()
                    progress = int(self._now) - now
                    if progress > 0:
                        pbar.update(progress)
            else:
                while True:
                    self.step()
        except StopSimulation as exc:
            if show_progress_bar:
                pbar.update(total - pbar.n)
                pbar.close()
            return exc.args[0]  # == until.value
        except EmptySchedule as e:
            if until is not None:
                assert not until.triggered  # noqa: S101
                msg = f'No scheduled events left but "until" event was not triggered: {until}'
                raise RuntimeError(msg) from e
        return None



from dataclasses import dataclass, field

import pytest

from cosimtlk.simulation.state import SimulationState


@dataclass
class SubState:
    i: int = field(default=1)
    f: float = field(default=1.0)
    s: str = field(default="sub")


@dataclass
class State(SimulationState):
    i: int = field(default=1)
    f: float = field(default=1.0)
    s: str = field(default="main")
    ss: SubState = field(default_factory=SubState)


@pytest.fixture(scope="function")
def state():
    return State()


def test_setattr_getattr(state):
    assert state.i == 1
    state.i = 2
    assert state.i == 2


def test_setattr_getattr_subclass(state):
    assert state.ss.i == 1
    state.ss.i = 2
    assert state.ss.i == 2


def test_setitem_getitem(state):
    assert state["i"] == 1
    state["i"] = 2
    assert state["i"] == 2


def test_setitem_getitem_subclass(state):
    assert state["ss.i"] == 1
    state["ss.i"] = 2
    assert state["ss.i"] == 2


def test_set_get_single(state):
    state.set(i=2)
    assert state.i == 2


def test_set_get_multiple(state):
    state.set(i=2, f=2.0)
    assert state.i == 2
    assert state.f == 2


def test_set_get_multiple_subclass(state):
    state.set(**{"ss.i": 2, "ss.f": 2.0})
    assert state.ss.i == 2
    assert state.ss.f == 2


if __name__ == "__main__":
    pytest.main()



from typing import Any


class SimulationState:
    def __setitem__(self, key, value):
        if "." in key:
            keys = key.split(".")
            current = self
            for k in keys[:-1]:
                current = getattr(current, k)
            setattr(current, keys[-1], value)
        else:
            setattr(self, key, value)

    def __getitem__(self, key):
        if "." in key:
            keys = key.split(".")
            current = self
            for k in keys:
                current = getattr(current, k)
            return current
        return getattr(self, key)

    def set(self, **states: Any) -> None:  # noqa: A003
        for key, value in states.items():
            self.__setitem__(key, value)



import pytest
from fmpy.model_description import ModelDescription

from cosimtlk import FMU, FMUInstance


def test_can_instantiate():
    FMU("tests/fixtures/fmus/ModSim.Examples.InputTest.fmu")


def test_model_description(local_fmu):
    assert isinstance(local_fmu.model_description, ModelDescription)


def test_inputs(local_fmu):
    for input_ in local_fmu.inputs:
        assert input_.causality == "input"


def test_outputs(local_fmu):
    for output in local_fmu.outputs:
        assert output.causality == "output"


def test_parameters(local_fmu):
    for parameter in local_fmu.parameters:
        assert parameter.causality == "parameter"


def test_instantiate(local_fmu):
    fmu = local_fmu.instantiate(
        start_time=0,
        step_size=1,
        start_values={},
    )
    assert isinstance(fmu, FMUInstance)


if __name__ == "__main__":
    pytest.main()



import logging
from typing import Annotated

import typer
from rich.logging import RichHandler

app = typer.Typer(
    name="cosimtlk",
    help="A tool to simulate FMUs.",
    add_completion=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.callback()
def main(
    verbose: bool = False,  # noqa
) -> None:
    if verbose:
        logging.basicConfig(level=logging.DEBUG, handlers=[RichHandler()])
    else:
        logging.basicConfig(level=logging.INFO, handlers=[RichHandler()])


@app.command()
def server(
    host: Annotated[str, typer.Option("--host", "-h", help="Host to listen on.")] = "127.0.0.1",
    port: Annotated[int, typer.Option("--port", "-p", help="Port to listen on.")] = 8000,
    reload: Annotated[bool, typer.Option("--reload", "-r", help="Reload server on file changes.")] = False,  # noqa
):
    try:
        import uvicorn
    except ImportError:
        typer.echo("Please install uvicorn to run the server.")
    else:
        uvicorn.run("cosimtlk.app.main:app", host=host, port=port, reload=reload)



from collections.abc import Callable, Generator

from pandas import DataFrame, Series

from cosimtlk.simulation.entities import Entity


class Input(Entity):
    def __init__(self, name: str, *, values: Series):
        """An entity that sets the input of the simulator based on the input date.

        Args:
            name: The name of the entity.
            values: The input data as a Series with a DatetimeIndex.
                The index of the values is used as the time at which the input is set
        """
        super().__init__(name)
        self.values = values
        self._state = values.name
        self._index = 0

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.set_inputs_process]

    def set_inputs_process(self):
        while True:
            if self.values.empty or self._index >= len(self.values):
                self.log.warning("no data left!")
                break

            current_time = self.ctx.current_datetime
            next_point_at = self.values.index[self._index]

            if next_point_at <= current_time:
                current_value = self.values.iloc[self._index]
                self.log.debug(f"setting {self._state}={current_value}")
                self.ctx.state[self._state] = current_value
                self._index += 1
            else:
                next_point_in = int((next_point_at - current_time).total_seconds())
                yield self.wait_for(next_point_in)


class MultiInput(Entity):
    def __init__(self, name: str, *, values: DataFrame):
        """An entity that sets the input of the simulator based on the input date.

        Args:
            name: The name of the entity.
            values: The input data as a Series with a DatetimeIndex.
                The index of the values is used as the time at which the input is set
        """
        super().__init__(name)
        self.values = values
        self._index = 0

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.set_inputs_process]

    def set_inputs_process(self):
        while True:
            if self.values.empty or self._index >= len(self.values):
                self.log.warning("no data left!")
                break

            current_time = self.ctx.current_datetime
            next_point_at = self.values.index[self._index]

            if next_point_at <= current_time:
                current_values = self.values.iloc[self._index]
                self.log.debug(f"setting {current_values.to_dict()}")
                self.ctx.state.set(**current_values)
                self._index += 1
            else:
                next_point_in = int((next_point_at - current_time).total_seconds())
                yield self.wait_for(next_point_in)



from __future__ import annotations

import functools
from collections.abc import Callable, Generator
from dataclasses import dataclass, field

from cosimtlk.simulation.entities import Entity


@dataclass(frozen=True, slots=True)
class Measurement:
    name: str
    store_as: str | None = field(default=None)

    def __post_init__(self):
        object.__setattr__(self, "store_as", self.store_as or self.name)


class StateObserver(Entity):
    def __init__(
        self,
        name: str,
        *,
        measurements: list[Measurement],
        scheduler: Callable,
    ):
        """An entity that stores the current state of the simulation into long term storage.

        Args:
            name: The name of the entity.
            measurements: A mapping of state names to measurement names. The keys are used as
                the names of the measurements inside the database, while the values determine the state.
            scheduler: A generator function that schedules a function such as `cosimtlk.simulation.utils.every`
                or `cosimtlk.simulation.utils.cron`.
        """
        super().__init__(name)
        self.measurements = measurements
        self.scheduler = scheduler

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        scheduled_process = self.scheduler(self.__class__.sensing_process)
        return [
            functools.partial(scheduled_process, self),
        ]

    def sensing_process(self):
        values = {measurement.store_as: self.ctx.state[measurement.name] for measurement in self.measurements}
        self.log.debug(f"observed measurements={values}")
        self.ctx.db.store_observations(self.ctx.current_datetime, **values)
